/*
 * LZW compression from memory to memory
 * Description of the LZW algorithm:
 * http://dogma.net/markn/articles/lzw/lzw.htm
 */

#include "lzw.h"
#include <stdlib.h>

typedef void (*WRITE_COMP)(int codeword, int *destination, long *position);

typedef struct _lzwDictionaryEntry
{
    unsigned char root;
    int codeword;
    unsigned char occupied;
} lzwDictionaryEntry;

long generalizedCompress(WRITE_COMP outFunc, unsigned char *uncompressedMem, int *encodedMem, long uncompressedSize);
unsigned char getNextUncompressedByte(long *bytesRead, unsigned char *uncompressedMem);
void outputCodeword(int codeword, int *destination, long *codewordsWritten);
void discardCodeword(int codeword, int *destination, long *codewordsWritten);

int getHashCode(unsigned char root, int codeword, lzwDictionaryEntry* dictionary);
int probe1(unsigned char root, int codeword);
int probe2(unsigned char root, int codeword);
int probe3(int hashCode);
unsigned char hashPosFound(int hashCode, unsigned char root, int codeword, lzwDictionaryEntry* dictionary);

long lzwGetCodewordCount(unsigned char *uncompressedMem, long uncompressedSize)
{
    return(generalizedCompress(&discardCodeword, uncompressedMem, NULL, uncompressedSize));
}

long lzwCompress(unsigned char *uncompressedMem, int *encodedMem, long uncompressedSize)
{
    return(generalizedCompress(&outputCodeword, uncompressedMem, encodedMem, uncompressedSize));
}

/* --------------------------------------------------------------------------------------
   Functions used only inside lzw.c
   -------------------------------------------------------------------------------------- */

long generalizedCompress(WRITE_COMP outFunc, unsigned char *uncompressedMem, int *encodedMem, long uncompressedSize)
{
    int i;

    /* re-initialize the dictionary when there are more than 0xccc entries */
    const int maxDictEntries = 0xccc;

    const int lzwDictionarySize = 0x1000;   /* make the hash table 25% larger than needed */

    int currentString;
    unsigned char character;
    int hashCode;
    unsigned char c1,c2,c3;   /* pseudo-booleans */
    long bytesRead = 0;
    long codewordsWritten = 0;

    /* initialize the dictionary */
    lzwDictionaryEntry *lzwDictionary = (lzwDictionaryEntry *) malloc(sizeof(lzwDictionaryEntry[lzwDictionarySize]));
    int codewordsInDictionary = 0;

    /* clear the dictionary */
    memset(lzwDictionary, 0, sizeof(lzwDictionaryEntry[lzwDictionarySize]));
    for (i = 0; i < 0x100; i++)
    {
        lzwDictionary[i].occupied = 1;
    }

    if (bytesRead + 1 <= uncompressedSize)   /* there's at least 1 byte left in the input */
    {
        /* STRING = get input character */
        currentString = getNextUncompressedByte(&bytesRead, uncompressedMem);

        /* WHILE there are still input characters DO */
        while (bytesRead + 1 <= uncompressedSize)
        {
            /* CHARACTER = get input character */
            character = getNextUncompressedByte(&bytesRead, uncompressedMem);

            /* is STRING+CHARACTER in the string table? */
            hashCode = getHashCode(character, currentString, lzwDictionary);

            if (lzwDictionary[hashCode].occupied == 1)
            {
                /* STRING+CHARACTER is in the string table */

                /* STRING = STRING+character */
                currentString = hashCode;
            }
            else
            {
                /* STRING+CHARACTER has yet to be added */

                /* output the code for STRING */
                outFunc(currentString, encodedMem, &codewordsWritten);
                
                /* add STRING+CHARACTER to the string table */
                lzwDictionary[hashCode].occupied = 1;
                lzwDictionary[hashCode].root = character;
                lzwDictionary[hashCode].codeword = currentString;
                codewordsInDictionary++;

                /* STRING = CHARACTER */
                currentString = character;

                /* check if the dictionary is full */
                if (codewordsInDictionary > maxDictEntries)
                {
                    /* output the code for STRING */
                    outFunc(currentString, encodedMem, &codewordsWritten);

                    /* clear the dictionary */
                    memset(lzwDictionary, 0, sizeof(lzwDictionaryEntry[lzwDictionarySize]));
                    for (i = 0; i < 0x100; i++)
                    {
                        lzwDictionary[i].occupied = 1;
                    }
                    codewordsInDictionary = 0;

                    if (bytesRead + 1 <= uncompressedSize)   /* there's at least 1 byte left in the input */
                    {
                        /* STRING = get input character */
                        currentString = getNextUncompressedByte(&bytesRead, uncompressedMem);
                    }
                    else
                    {
                        return(codewordsWritten);
                    }
                }
            }
        }

        /* output the code for STRING */
        outFunc(currentString, encodedMem, &codewordsWritten);
    }

    return(codewordsWritten);
}

unsigned char getNextUncompressedByte(long *bytesRead, unsigned char *uncompressedMem)
{
    unsigned char uncompressedByte = uncompressedMem[*bytesRead];
    (*bytesRead)++;
    return(uncompressedByte);
}

void outputCodeword(int codeword, int *destination, long *codewordsWritten)
{
    destination[*codewordsWritten] = codeword;
    (*codewordsWritten)++;
}

void discardCodeword(int codeword, int *destination, long *codewordsWritten)
{
    (*codewordsWritten)++;
}

int getHashCode(unsigned char root, int codeword, lzwDictionaryEntry* dictionary)
{
    int hashCode;

    /* probe 1 */
    hashCode = probe1(root, codeword);
    if (hashPosFound(hashCode, root, codeword, dictionary)) {
        return(hashCode);
    }
    /* probe 2 */
    hashCode = probe2(root, codeword);
    if (hashPosFound(hashCode, root, codeword, dictionary)) {
        return(hashCode);
    }
    /* probe 3 */
    do {
        hashCode = probe3(hashCode);
    }
    while (! hashPosFound(hashCode, root, codeword, dictionary));

    return(hashCode);
}

int probe1(unsigned char root, int codeword)
{
    int newHashCode = ((root << 4) ^ codeword) & 0xfff;
    return(newHashCode);
}

/* The secondary probe uses some assembler instructions that aren't easily translated to C. */
int probe2(unsigned char root, int codeword)
{
    /* registers[0] == AX, registers[1] == DX */
    long registers[2], temp;
    long carry, oldCarry;
    int i,j;

    /* the pre-mul part */
    registers[1] = 0;
    registers[0] = ((root << 1) + codeword) | 0x800;

    /* the mul part (simulated mul instruction) */
    /* DX:AX = AX * AX                          */
    temp = (registers[0] & 0xff) * (registers[0] & 0xff);
    temp += 2 * (registers[0] & 0xff) * (registers[0] >> 8) * 0x100;
    registers[1] = (temp >> 16) + (registers[0] >> 8) * (registers[0] >> 8);
    registers[0] = temp & 0xffff;

    /* if DX != 0, the mul instruction sets the carry flag */
    if (registers[1] == 00) {carry = 0;}
    else {carry = 1;}

    /* the rcl part */
    for (i = 0; i < 2; i++)   /* 2 rcl's */
    {
        for (j = 0; j < 2; j++)   /* rotate through 2 registers */
        {
            oldCarry = carry;
            carry = (registers[j] >> 15) & 1;
            registers[j] = (registers[j] << 1) | oldCarry;
            registers[j] = registers[j] & 0xffff;   /* make sure register stays 16 bit */
        }        
    }

    /* final touches */
    registers[0] = ((registers[0] >> 8) | (registers[1] << 8)) & 0xfff;

    return((int)registers[0]);
}

int probe3(int hashCode)
{
    const long probeOffset = 0x1fd;   /* I think 0x1fd is prime */

    long newHashCode = (hashCode + probeOffset) & 0xfff;
    return((int)newHashCode);
}

unsigned char hashPosFound(int hashCode, unsigned char root, int codeword, lzwDictionaryEntry* dictionary)
{
    if (hashCode > 0xff)   /* hash codes must not be roots */
    {
        unsigned char c1, c2, c3;

        if (dictionary[hashCode].occupied)
        {
            /* hash table position is occupied */
            c1 = 1;
            /* is our (root,codeword) pair already in the hash table? */
            c2 = dictionary[hashCode].root == root;
            c3 = dictionary[hashCode].codeword == codeword;
        }
        else
        {
            /* hash table position is free */
            c1 = 0;
        }

        return((!c1) || (c1 && c2 && c3));
    }
    else
    {
        return(0);
    }
}
